/*
 * Copyright © 2020 camunda services GmbH (info@camunda.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.atomix.raft.impl;

import static org.assertj.core.api.Assertions.assertThat;

import io.atomix.utils.concurrent.SingleThreadContext;
import java.time.Duration;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

final class PriorityElectionTimerTest {

  private final Logger log = LoggerFactory.getLogger(PriorityElectionTimerTest.class);
  private final SingleThreadContext threadContext = new SingleThreadContext("priorityElectionTest");

  @AfterEach
  void afterEach() {
    threadContext.close();
  }

  @Test
  void shouldLowerPriorityNodeEventuallyStartsAnElection() {
    // given
    final AtomicInteger triggerCount = new AtomicInteger();

    final PriorityElectionTimer timer =
        new PriorityElectionTimer(
            Duration.ofMillis(100), threadContext, triggerCount::getAndIncrement, log, 4, 1);

    // when
    timer.reset();

    // then
    Awaitility.await().until(() -> triggerCount.get() >= 1);
  }

  @Test
  void shouldHighPriorityNodeStartElectionFirst() {
    // given
    final String highPrioId = "highPrioTimer";
    final String lowPrioId = "lowPrioTimer";
    final List<String> electionOrder = new CopyOnWriteArrayList<>();

    final int targetPriority = 4;
    final PriorityElectionTimer timerHighPrio =
        new PriorityElectionTimer(
            Duration.ofMillis(100),
            threadContext,
            () -> electionOrder.add(highPrioId),
            log,
            targetPriority,
            targetPriority);

    final PriorityElectionTimer timerLowPrio =
        new PriorityElectionTimer(
            Duration.ofMillis(100),
            threadContext,
            () -> electionOrder.add(lowPrioId),
            log,
            targetPriority,
            1);

    // when
    timerLowPrio.reset();
    timerHighPrio.reset();

    // then
    Awaitility.await()
        .untilAsserted(
            () ->
                assertThat(electionOrder)
                    .as("both elections should have been triggered eventually")
                    .contains(highPrioId, lowPrioId));
    assertThat(electionOrder.get(0))
        .as("the first election triggered should have been the high priority election")
        .isEqualTo(highPrioId);
  }
}
